package ru.dbotthepony.kstarbound.world

import it.unimi.dsi.fastutil.objects.Object2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import ru.dbotthepony.kbox2d.api.ContactImpulse
import ru.dbotthepony.kbox2d.api.IContactFilter
import ru.dbotthepony.kbox2d.api.IContactListener
import ru.dbotthepony.kbox2d.api.Manifold
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kbox2d.dynamics.B2World
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import ru.dbotthepony.kstarbound.METRES_IN_STARBOUND_UNIT
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.util.Timer
import ru.dbotthepony.kstarbound.world.entities.CollisionResolution
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kstarbound.world.entities.projectile.AbstractProjectileMovementController
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.util2d.AABBi
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.nint.Vector2i
import kotlin.math.absoluteValue

const val EARTH_FREEFALL_ACCELERATION = 9.8312 / METRES_IN_STARBOUND_UNIT

data class WorldSweepResult(
	val hitPosition: Vector2d,
	val collisionTime: Double,
	val hitAnything: Boolean,
	val hitNormal: Vector2d
)

private const val EPSILON = 0.00001

abstract class World<This : World<This, ChunkType>, ChunkType : Chunk<This, ChunkType>>(
	val seed: Long,
	val widthInChunks: Int
) {
	protected val chunkMap = Object2ObjectAVLTreeMap<ChunkPos, ChunkType> cmp@{ a, b ->
		return@cmp a.compareTo(b)
	}

	/**
	 * Является ли мир "сферическим"
	 *
	 * Данный флаг говорит о том, что [widthInChunks] имеет осмысленное значение,
	 * попытка получить чанк с X координатой меньше нуля или больше [widthInChunks]
	 * приведёт к замыканию на конец/начало мира соответственно
	 */
	val isCircular = widthInChunks > 0

	/**
	 * Chunks, which have their collision mesh changed
	 */
	val dirtyPhysicsChunks = HashSet<ChunkType>()

	protected var lastAccessedChunk: ChunkType? = null

	val physics = B2World(Vector2d(0.0, -EARTH_FREEFALL_ACCELERATION))

	private var timers = ArrayList<Timer>()

	fun timer(period: Double, executionTimes: Int, func: (Timer) -> Unit): Timer {
		val timer = Timer(period, executionTimes, func)
		timers.add(timer)
		return timer
	}

	init {
		physics.contactFilter = object : IContactFilter {
			override fun shouldCollide(fixtureA: B2Fixture, fixtureB: B2Fixture): Boolean {
				val dataA = fixtureA.body!!.userData
				val dataB = fixtureB.body!!.userData

				if (dataA is AbstractProjectileMovementController && dataB is AbstractProjectileMovementController) {
					return false
				}

				return true
			}
		}

		physics.contactListener = object : IContactListener {
			override fun beginContact(contact: AbstractContact) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.beginContact(contact)
				}

				if (dataB is IContactListener) {
					dataB.beginContact(contact)
				}
			}

			override fun endContact(contact: AbstractContact) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.endContact(contact)
				}

				if (dataB is IContactListener) {
					dataB.endContact(contact)
				}
			}

			override fun preSolve(contact: AbstractContact, oldManifold: Manifold) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.preSolve(contact, oldManifold)
				}

				if (dataB is IContactListener) {
					dataB.preSolve(contact, oldManifold)
				}
			}

			override fun postSolve(contact: AbstractContact, impulse: ContactImpulse) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.postSolve(contact, impulse)
				}

				if (dataB is IContactListener) {
					dataB.postSolve(contact, impulse)
				}
			}
		}
	}

	/**
	 * Таймер этого мира, в секундах.
	 *
	 * Условия:
	 * * Это значение МОЖЕТ уменьшаться, в том числе в силу того, что клиент
	 * обсчитывает мир быстрее, чем сервер, и наоборот.
	 * * Точка отсчёта не имеет смысла (по умолчанию отсчитывается от нуля).
	 * * Все сущности обязаны использовать этот таймер для определения времени своих событий
	 * * Таймер НЕ МОЖЕТ возвращать разные значения в пределах одного тика.
	 * т.е. таймер должен быть обновлён только после или перед тиком всех потомков
	 * * НЕ ОБЯЗАН соответствовать течению реального времени, но как правило, соответствует ему
	 * (если позволяет скорость обсчёта потомков).
	 * * ОБЯЗАН быть сохранён при сохранении мира, и выставляться при его загрузке.
	 */
	var timer: Double = 0.0
		protected set

	/**
	 * Заставляет этот мир "думать" на заданное количество секунд.
	 * Вызывает все необходимые методы у потомков, обновляя игровое состояние.
	 */
	fun think(delta: Double) {
		require(delta > 0.0) { "Tried to update $this by $delta seconds" }

		try {
			for (chunk in dirtyPhysicsChunks) {
				chunk.bakeCollisions()
			}

			dirtyPhysicsChunks.clear()

			physics.step(delta, 6, 4)

			timer += delta

			if (timers.isNotEmpty()) {
				val current = timers
				timers = ArrayList()

				val iterator = current.iterator()

				for (timer in iterator) {
					timer.think(delta)

					if (timer.destroyed) {
						iterator.remove()
					}
				}

				if (timers.isNotEmpty()) {
					current.addAll(timers)
				}

				timers = current
			}

			thinkInner(delta)
		} catch(err: Throwable) {
			throw RuntimeException("Ticking world $this", err)
		}
	}

	protected abstract fun thinkInner(delta: Double)

	/**
	 * Сущности, у которых нет чанка, но они находятся в этом мире
	 */
	val orphanedEntities = HashSet<Entity>()

	/**
	 * Сущности, находящиеся в этом мире
	 */
	val entities = HashSet<Entity>()

	/**
	 * Стандартное ускорение свободного падения в Starbound Units/секунда^2
	 *
	 * При Vector2d.ZERO = невесомость
	 */
	var gravity by physics::gravity

	protected abstract fun chunkFactory(
		pos: ChunkPos,
	): ChunkType

	/**
	 * Возвращает чанк на указанной позиции если он существует
	 */
	open operator fun get(pos: ChunkPos): ChunkType? {
		if (!isCircular) {
			val lastAccessedChunk = lastAccessedChunk

			if (lastAccessedChunk?.pos == pos) {
				return lastAccessedChunk
			}

			return chunkMap[pos]
		}

		@Suppress("Name_Shadowing")
		val pos = pos.circular(widthInChunks)

		val lastAccessedChunk = lastAccessedChunk

		if (lastAccessedChunk?.pos == pos) {
			return lastAccessedChunk
		}

		val load = chunkMap[pos]
		this.lastAccessedChunk = load
		return load
	}

	open fun getInstantTuple(pos: ChunkPos): IWorldChunkTuple<This, ChunkType>? {
		return this[pos]?.let { InstantWorldChunkTuple(this as This, it) }
	}

	/**
	 * Возвращает чанк на заданной позиции, создаёт его если он не существует
	 */
	open fun computeIfAbsent(pos: ChunkPos): ChunkType {
		@Suppress("Name_Shadowing")
		val pos = if (isCircular) pos.circular(widthInChunks) else pos

		val _lastAccessedChunk = lastAccessedChunk

		if (_lastAccessedChunk?.pos == pos) {
			return _lastAccessedChunk
		}

		return chunkMap.computeIfAbsent(pos, Object2ObjectFunction {
			val chunk = chunkFactory(pos)

			lastAccessedChunk = chunk

			val orphanedInThisChunk = ArrayList<Entity>()

			for (ent in orphanedEntities) {
				val cPos = if (isCircular) ChunkPos.fromTilePosition(ent.position, widthInChunks) else ChunkPos.fromTilePosition(ent.position)

				if (cPos == pos) {
					orphanedInThisChunk.add(ent)
				}
			}

			for (ent in orphanedInThisChunk) {
				ent.chunk = chunk
			}

			return@Object2ObjectFunction chunk
		})
	}

	open fun getForegroundView(pos: ChunkPos): TileView? {
		val get = get(pos) ?: return null
		val tuple = InstantWorldChunkTuple(this as This, get)

		return TileView(
			center = 			tuple.chunk.foreground,
			left = 				tuple.left?.chunk?.foreground,
			top = 				tuple.top?.chunk?.foreground,
			topLeft = 			tuple.topLeft?.chunk?.foreground,
			topRight = 			tuple.topRight?.chunk?.foreground,
			right = 			tuple.right?.chunk?.foreground,
			bottom = 			tuple.bottom?.chunk?.foreground,
			bottomLeft = 		tuple.bottomLeft?.chunk?.foreground,
			bottomRight = 		tuple.bottomRight?.chunk?.foreground,
		)
	}

	open fun getBackgroundView(pos: ChunkPos): TileView? {
		val get = get(pos) ?: return null
		val tuple = InstantWorldChunkTuple(this as This, get)

		return TileView(
			center = 			tuple.chunk.background,
			left = 				tuple.left?.chunk?.background,
			top = 				tuple.top?.chunk?.background,
			topLeft = 			tuple.topLeft?.chunk?.background,
			topRight = 			tuple.topRight?.chunk?.background,
			right = 			tuple.right?.chunk?.background,
			bottom = 			tuple.bottom?.chunk?.background,
			bottomLeft = 		tuple.bottomLeft?.chunk?.background,
			bottomRight = 		tuple.bottomRight?.chunk?.background,
		)
	}

	fun getTile(pos: Vector2i): ITileState? {
		return get(ChunkPos.fromTilePosition(pos))?.foreground?.get(ChunkPos.normalizeCoordinate(pos.x), ChunkPos.normalizeCoordinate(pos.y))
	}

	fun getBackgroundTile(pos: Vector2i): ITileState? {
		return get(ChunkPos.fromTilePosition(pos))?.background?.get(ChunkPos.normalizeCoordinate(pos.x), ChunkPos.normalizeCoordinate(pos.y))
	}

	/**
	 * Возвращает все чанки, которые пересекаются с заданным [boundingBox]
	 */
	open fun collect(boundingBox: AABBi): List<ChunkType> {
		val output = ArrayList<ChunkType>()

		for (pos in boundingBox.chunkPositions) {
			val chunk = get(pos)

			if (chunk != null) {
				output.add(chunk)
			}
		}

		return output
	}

	/**
	 * Возвращает все чанки, которые пересекаются с заданным [boundingBox]
	 */
	open fun collectPositionAware(boundingBox: AABBi): List<Pair<ChunkPos, ChunkType>> {
		val output = ArrayList<Pair<ChunkPos, ChunkType>>()

		for (pos in boundingBox.chunkPositions) {
			val chunk = get(pos)

			if (chunk != null) {
				output.add(pos to chunk)
			}
		}

		return output
	}

	/**
	 * Производит проверку на пересечение [worldaabb] с геометрией мира при попытке пройти в [_deltaMovement]
	 */
	fun sweep(
		worldaabb: AABB,
		_deltaMovement: Vector2d,
		collisionResolution: CollisionResolution,
		delta: Double
	): WorldSweepResult {
		var deltaMovement = _deltaMovement

		val potentialAABB = worldaabb + deltaMovement
		val combined = worldaabb.combine(potentialAABB)

		val collected = collect((combined).encasingChunkPosAABB())
			.map { it.foreground.collisionLayers() }
			.flatten()
			.sortedWith { o1, o2 ->
				val a = o1.distance(worldaabb)
				val b = o2.distance(worldaabb)

				if (a == b)
					return@sortedWith 0

				if (a > b)
					return@sortedWith 1

				return@sortedWith -1
			}

		var hitAnything = false
		var hitNormal = Vector2d.ZERO

		if (collected.isNotEmpty()) {
			for (iteration in 0 .. 100) {
				var collided = false

				for (aabb in collected) {
					// залез в блоки?
					/*if (potentialAABB.intersectWeak(aabb)) {
						val push = worldaabb.pushOutFrom(aabb)
						println("into tiles")

						if (push.length > 0.0) {
							if (iteration == 0) {
								collided = true
								continue
							} else {
								hitAnything = true
								deltaMovement -= push * delta * 5.0
								potentialAABB = worldaabb + deltaMovement
								combined = worldaabb.combine(potentialAABB)

								collided = true
								continue
							}
						}
					}*/

					// ранний тест (отсечение заведомо не пересекаемой геометрии)
					if (!aabb.intersect(combined)) {
						continue
					}

					// обычный тест коллизии
					val (normal, collisionTime) = worldaabb.sweep(aabb, deltaMovement)

					if (collisionTime != 1.0) {
						hitAnything = true
						collided = true
						hitNormal = normal

						if (weakDoubleZeroing(collisionTime) == 0.0) {
							// Мы ВПЛОТНУЮ стоим к другой геометрии (пересекаемся одной из осей)
							deltaMovement = Vector2d.ZERO
						} else {
							val remainingTime = 1.0 - collisionTime

							when (collisionResolution) {
								CollisionResolution.STOP -> {
									// val weak = weakDoubleZeroing(collisionTime, epsilon = EPSILON)

									for (i in 100 downTo 1) {
										val move = deltaMovement * collisionTime * (i / 100.0)

										// сегодня, в программе
										// погрешность плавающей запятой
										if (!(worldaabb + move).intersect(aabb)) {
											deltaMovement = move
											break
										} // ЭтО чО вАщЕ ТакОе?
									}
								}

								CollisionResolution.PUSH -> {
									var dot = deltaMovement.pseudoDot(normal)
									val magnitude = deltaMovement.length * weakDoubleZeroing(remainingTime, epsilon = EPSILON)

									if (dot > 0.0) {
										dot = 1.0
									} else {
										dot = -1.0
									}

									deltaMovement = Vector2d(dot * normal.y * magnitude, dot * normal.x * magnitude)
								}

								CollisionResolution.SLIDE -> {
									val dot = deltaMovement.pseudoDot(normal) * remainingTime

									if (weakDoubleZeroing(dot) != 0.0) {
										deltaMovement = Vector2d(dot * normal.y, dot * normal.x)
									} else {
										// Граничный случай: Направления идеально противоположны.
										// Обрабатываем как CollisionResolution.STOP

										// val weak = weakDoubleZeroing(collisionTime, epsilon = EPSILON)

										for (i in 100 downTo 1) {
											val move = deltaMovement * collisionTime * (i / 100.0)

											// сегодня, в программе
											// погрешность плавающей запятой
											if (!(worldaabb + move).intersect(aabb)) {
												deltaMovement = move
												break
											} // ЭтО чО вАщЕ ТакОе?
										}
									}
								}

								CollisionResolution.BOUNCE -> {
									deltaMovement *= weakDoubleZeroing(remainingTime, epsilon = EPSILON)

									if (normal.x.absoluteValue > 0.00001 && normal.y.absoluteValue > 0.00001) {
										deltaMovement *= Vector2d.NEGATIVE_XY
									} else if (normal.x.absoluteValue > 0.00001) {
										deltaMovement *= Vector2d.NEGATIVE_X
									} else if (normal.y.absoluteValue > 0.00001) {
										deltaMovement *= Vector2d.NEGATIVE_Y
									}
								}
							}
						}
					}
				}

				if (!collided) {
					//println("resolved in $iteration")
					break
				}
			}
		}

		return WorldSweepResult(
			hitAnything = hitAnything,
			hitNormal = hitNormal,
			collisionTime = worldaabb.centre.distance(deltaMovement) / worldaabb.centre.distance(_deltaMovement),
			hitPosition = deltaMovement,
		)
	}

	/**
	 * Возвращает, застрянет ли сущность будет с коллизией [worldaabb]
	 */
	fun isSpaceEmptyFromTiles(worldaabb: AABB): Boolean {
		val collected = collect(worldaabb.encasingChunkPosAABB())
			.map { it.foreground.collisionLayers() }
			.flatten()
			.sortedWith { o1, o2 ->
				val a = o1.distance(worldaabb)
				val b = o2.distance(worldaabb)

				if (a == b)
					return@sortedWith 0

				if (a > b)
					return@sortedWith 1

				return@sortedWith -1
			}

		if (collected.isEmpty()) {
			return true
		}

		for (aabb in collected) {
			if (aabb.intersectWeak(worldaabb)) {
				return false
			}
		}

		return true
	}
}
